/*  
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)
    
    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 
*/

#include "filter.h"
#include "filtprog.h"
#include "cache.h"

#ifndef PIPE_BUF
#define PIPE_BUF 4096
#endif

#define FBUFSIZ 1024

typedef struct {
  const char *const *prog;
  int isgzip;
  int major;
} filt_devd;

struct filter {
  cache_obj *cobj;
  int fh;
  avoff_t size;
  filt_devd *dd;
  void *filtpr;
  struct avstat stbuf;
};

typedef struct {
  struct filter *filt;
  avoff_t ptr;
} filt_fdi;


static avoff_t get_gzip_size(ave *v, int fh)
{
  avoff_t origptr;
  unsigned char nbuf[4];
  avoff_t size;
  unsigned char buf[2];
  avssize_t rres;
  
  origptr = __av_lseek(v, fh, 0, AVSEEK_CUR);
  if(origptr == -1) return -1;

  size = -1;
  if(__av_lseek(v, fh, 0, AVSEEK_SET) == -1) goto end;

  rres = __av_read(v, fh, buf, 2);
  if(rres == -1) goto end;
    
  if(rres == 2 && buf[0] == 0x1f && buf[1] == 0x8b) {
    if(__av_lseek(v, fh, -4, AVSEEK_END) != -1 &&
       __av_read(v, fh, nbuf, 4) == 4) {
      size = nbuf[0] + (nbuf[1] << 8) + (nbuf[2] << 16) + (nbuf[3] << 24);
    }
  }
  else size = -2; /* Meaning we don't know */
  
 end:
  if(__av_lseek(v, fh, origptr, AVSEEK_SET) == -1) return -1;
  return size;
}


static int get_size(ave *v, filt_fdi *di)
{
  avoff_t res;
  
  if(di->filt->dd->isgzip) {
    res = get_gzip_size(v, di->filt->fh);
    if(res == -1) return -1;
    if(res >= 0) {
      di->filt->size = res;
      return 0;
    }
  }
  res = __av_filt_getsize(v, di->filt->filtpr);
  if(res == -1) return -1;
  
  di->filt->size = res;

  return 0;
}

static avssize_t orig_read(ave *v, void *devinfo, char *buf, avsize_t nbyte)
{
  struct filter *filt = (struct filter *) devinfo;
  return __av_read(v, filt->fh, buf, nbyte);
}

static avoff_t orig_lseek(ave *v, void *devinfo, avoff_t offset, int whence)
{
  struct filter *filt = (struct filter *) devinfo;
  return __av_lseek(v, filt->fh, offset, whence);
}

static void filter_free(void *obj)
{
  struct filter *filt = (struct filter *) obj;

  if(filt->filtpr != AVNULL) __av_close_filtprog(DUMMYV, filt->filtpr);
  if(filt->fh != -1) __av_close(DUMMYV, filt->fh);

  __av_free(filt);
}

static int params_ok(ave *v, const char *params)
{
  if(params[0]) {
    if(params[0] == DIR_SEP_CHAR) v->errn = ENOTDIR;
    else v->errn = ENOENT;
    
    return 0;
  }
  return 1;
}

static int filt_valid(cache_obj *cobj, void *param)
{    
  struct filter *filt = (struct filter *) cobj->obj;
  avtime_t *mtimep = (avtime_t *) param;

  if(filt->stbuf.mtime != *mtimep) return -1;
  return 1;
}

static struct filter *get_filter(ave *v, void *devdata, vpath *path)
{
  filt_devd *dd = (filt_devd *) devdata;
  cache_obj *cobj;
  struct filt_params fp;
  struct avstat origbuf;
  struct filter *filt;
  int minor;

  if(__av_stat(v, BASE(path), &origbuf, AVFS_FLAG_NOSIZE) == -1) 
    return AVNULL;
  
  minor = __av_getminor(dd->major, origbuf.dev, origbuf.ino);
  cobj = __av_cache_find(v, dd->major, minor, filt_valid, 
			 (void *) &origbuf.mtime);

  if(cobj == AVNULL) return AVNULL;
  if(cobj->obj != AVNULL) return (struct filter *) cobj->obj;

  AV_NEW(v, filt);
  if(filt == AVNULL) {
    __av_cache_op(cobj, COBJ_FREE);
    return AVNULL;
  }
  
  cobj->obj = (void *) filt;
  cobj->destroy = filter_free;
  /* cobj->keep_time = ???; */

  filt->cobj = cobj;
  filt->fh = -1;
  filt->dd = dd;
  filt->size = -1;
  filt->filtpr = AVNULL;

  filt->stbuf = origbuf;
  filt->stbuf.mode &= ~(07000);  /* Clear SETU/GID bits  */
  filt->stbuf.blksize = PIPE_BUF;
  filt->stbuf.blocks  = 0;
  filt->stbuf.dev = __av_vmakedev(dd->major, minor);
  filt->stbuf.ino = ((minor >> 8) & 0xFFF) + 3;
  filt->stbuf.size = 0;
  
  filt->fh = __av_open(v, BASE(path), AVO_RDONLY, 0);
  if(filt->fh == -1) {
    __av_cache_op(cobj, COBJ_FREE);
    return AVNULL;
  }

  fp.prog = dd->prog;
  fp.readfunc = orig_read;
  fp.lseekfunc = orig_lseek;

  filt->filtpr = __av_open_filtprog(v, &fp, (void *) filt);
  if(filt->filtpr == AVNULL) {
    __av_cache_op(cobj, COBJ_FREE);
    return AVNULL;
  }
  
  __av_cache_op(cobj, COBJ_UNLOCK);
  return filt;
}

static void *filt_open(ave *v, void *devdata, vpath *path, 
		       int flags, int mode)
{
  filt_fdi *di = AVNULL;

  if(!params_ok(v, EXTRA(path))) return AVNULL;

  if((flags & AVO_ACCMODE) != AVO_RDONLY) {
    v->errn = EACCES;
    return AVNULL;
  }

  AV_NEW(v, di);
  if(di == AVNULL) return AVNULL;

  di->filt = get_filter(v, devdata, path);
  if(di->filt == AVNULL) {
    __av_free(di);
    return AVNULL;
  }
  di->ptr = 0;

  return (void *) di;
}

static int filt_close(ave *v, void *devinfo)
{
  filt_fdi *di = (filt_fdi *) devinfo;

  __av_cache_op(di->filt->cobj, COBJ_RELEASE);
  __av_free(di);

  return 0;
}


static avssize_t filt_read(ave *v, void *devinfo, char *buf, avsize_t nbyte)
{
  filt_fdi *di = (filt_fdi *) devinfo;
  avssize_t res;

  __av_cache_op(di->filt->cobj, COBJ_LOCK);  
  res = __av_filt_read(v, di->filt->filtpr, buf, di->ptr, nbyte);
  __av_cache_op(di->filt->cobj, COBJ_UNLOCK);  

  if(res != -1) di->ptr += res;
  
  return res;
}

static avssize_t filt_write(ave *v, void *devinfo, const char *buf,
			    avsize_t nbyte)
{
  v->errn = EBADF;
  return -1;
}

static avoff_t filt_lseek(ave *v, void *devinfo, avoff_t offset, int whence)
{
  filt_fdi *di = (filt_fdi *) devinfo;
  avoff_t size = 0;

  if(whence == AVSEEK_END) {
    int res = 0;

    __av_cache_op(di->filt->cobj, COBJ_LOCK);  
    if(di->filt->size == -1) res = get_size(v, di);
    size = di->filt->size;
    __av_cache_op(di->filt->cobj, COBJ_UNLOCK);  

    if(res == -1) return -1;
  }

  return __av_generic_lseek(v, &di->ptr, size, offset, whence);
}

static int filt_fstat(ave *v, void *devinfo, struct avstat *buf, int flags)
{
  filt_fdi *di = (filt_fdi *) devinfo;
  avoff_t size;
  int res = 0;

  *buf = di->filt->stbuf;

  __av_cache_op(di->filt->cobj, COBJ_LOCK);  
  if(di->filt->size == -1 && !(flags & AVFS_FLAG_NOSIZE)) 
    res = get_size(v, di);
  size = di->filt->size;
  __av_cache_op(di->filt->cobj, COBJ_UNLOCK);  
  
  if(res == -1) return -1;
  if(size != -1) buf->size = size;

  return 0;
}

static int filt_access(ave *v, void *devdata, vpath *path, int amode)
{
  if(!params_ok(v, EXTRA(path))) return -1;

  if(amode & (AVW_OK | AVX_OK)) {
    v->errn = EACCES;
    return -1;
  }
  
  return __av_access(v, BASE(path), amode);
}

static int filt_fgetfile(ave *v, void *devinfo, avoff_t offset, avsize_t *lenp,
			 avoff_t *retoffp)
{
  filt_fdi *di = (filt_fdi *) devinfo;
  int cachefd;
  int newfd;

  __av_cache_op(di->filt->cobj, COBJ_LOCK);  
  cachefd = __av_filt_getcachefile(v, di->filt->filtpr, offset + *lenp);
  if(cachefd != -1) newfd = __av_fgetfile(v, cachefd, offset, lenp, retoffp);
  else newfd = -1;
  __av_cache_op(di->filt->cobj, COBJ_UNLOCK);  
  
  return newfd;
}


static void filt_destroy(struct vdev_info *filt_vdev)
{
  filt_devd *dd = (filt_devd *) filt_vdev->devdata;

  if(dd != AVNULL) __av_cache_freeall(dd->major);

  __av_free(dd);
  __av_free(filt_vdev->name);
  __av_free(filt_vdev);
}

int __av_init_filt(ave *v, const char *name, const char *const prog[], 
		   struct ext_info *exts)
{
  filt_devd *dd;
  struct vdev_info *filt_vdev;
  int res;

  filt_vdev = __av_new_vdev(v);
  if(filt_vdev == AVNULL) return -1;

  filt_vdev->open     = filt_open;
  filt_vdev->close    = filt_close; 
  filt_vdev->read     = filt_read;
  filt_vdev->write    = filt_write;
  filt_vdev->lseek    = filt_lseek;
  filt_vdev->fstat    = filt_fstat;
  filt_vdev->access   = filt_access;

  filt_vdev->fgetfile = filt_fgetfile;
  filt_vdev->destroy  = filt_destroy;

  if((filt_vdev->name = __av_strdup(v, name)) == AVNULL) goto error;

  filt_vdev->exts = exts;
  
  filt_vdev->flags = 0;

  AV_NEW(v, dd);

  if((filt_vdev->devdata = (void *) dd) == AVNULL) goto error;
  dd->prog = AVNULL;

  dd->isgzip = 0;
  if(strcmp(name, "ugz") == 0) dd->isgzip = 1;

  dd->prog = prog;

  res = __av_add_vdev(v, filt_vdev);
  if(res == -1) goto error;

  dd->major = res;

  return res;

 error:
  filt_destroy(filt_vdev);
  return -1;
}
