/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    OP_FILE module (vfs layer)
*/

#include "op_util.h"

LOCAL3(read, avssize_t, int, char *, avsize_t)
LOCAL3(write, avssize_t, int, const char *, avsize_t)
LOCAL3(lseek, avoff_t, int, avoff_t, int)

static LOCAL2(access, int, const char *, int)
static LOCAL3(readlink, int, const char *, void *, size_t)

avssize_t __av_read(ave *v, int fh, char *buf, avsize_t nbyte)
{
  finfo *fi;
  ssize_t res;

  if(IS_LOCAL(fh)) return __av_localread(v, fh, buf, nbyte);

  fi = check_fh(v, fh);
  if(fi == NULL) return -1;

  if((fi->oflags & AVO_ACCMODE) == AVO_WRONLY) {
    v->errn = EBADF;
    return -1;
  }

  res = (*fi->vdev->read)(v, fi->devinfo, buf, nbyte);
  return res;
}

ssize_t __av_virtread(int fh, void *buf, size_t nbyte)
{
  ave v;
  ssize_t res;

  v.errn = errno;
  res =  __av_read(&v, fh, (char *) buf, nbyte);
  errno = v.errn;
  return res;
}

avssize_t __av_write(ave *v, int fh, const char *buf, avsize_t nbyte)
{
  finfo *fi;

  if(IS_LOCAL(fh)) return __av_localwrite(v, fh, buf, nbyte);

  fi = check_fh(v, fh);
  if(fi == NULL) return -1;

  if((fi->oflags & AVO_ACCMODE) == AVO_RDONLY) {
    v->errn = EBADF;
    return -1;
  }

  return (*fi->vdev->write)(v, fi->devinfo, buf, nbyte);
}


ssize_t __av_virtwrite(int fh, const void *buf, size_t nbyte)
{
  ave v;
  ssize_t res;

  v.errn = errno;
  res = __av_write(&v, fh, (const char *) buf, nbyte);
  errno = v.errn;
  return res;
}

static off_t do_lseek(ave *v, int fh, off_t offset, int whence)
{
  finfo *fi;

  if(IS_LOCAL(fh)) return __av_locallseek(v, fh, offset, whence);

  fi = check_fh(v, fh);
  if(fi == NULL) return -1;

  return (*fi->vdev->lseek)(v, fi->devinfo, offset, whence);
}

avoff_t __av_lseek(ave *v, int fh, avoff_t offset, int whence)
{
  if(DO_VIRTDIR(fh)) {
    finfo *fi;

    fi = &__av_priv.file_table[fh];
    switch(whence) {
    case SEEK_SET:
      if(fi->virtent->off != -1 && offset == fi->virtent->off) {
        fi->virtdir = 2;
        return offset;
      }
      else fi->virtdir = 1;
      break;
    case SEEK_CUR:
      if(fi->virtdir == 2 && offset >= 0) return fi->virtent->off + offset;
      else fi->virtdir = 1;

      break;
    case SEEK_END:
      if(fi->virtent->off != -1 && offset > -VIRTDIRSIZE) {
        fi->virtdir = 2;
        return fi->virtent->off;
      }
      else fi->virtdir = 1;
      break;
    }
  }

  return do_lseek(v, fh, offset, whence);
}

off_t __av_virtlseek(int fh, off_t offset, int whence)
{
  ave v;
  off_t res;

  v.errn = errno;
  AVLOGSYS("lseek", (whence == SEEK_CUR) ? "SEEK_CUR" :
              ((whence == SEEK_SET) ? "SEEK_SET" : "SEEK_END"),
              fh, offset);

  res = __av_lseek(&v, fh, offset, whence);

  AVLOGSYS("  (lseek)", "", res, v.errn);

  errno = v.errn;
  return res;
}


int __av_access(ave *v, vpath *path, int amode)
{
  int res;
  struct vdev_info *vdev = path->vdev;
  
  if(vdev == NULL) 
    res = __av_localaccess(v, path->ipath, amode);
  else {
    if(vdev->access != NULL)
      res = (*vdev->access)(v, vdev->devdata, path, amode);
    else {
      v->errn = ENODEV;
      res = -1;
    }
  }

  return res;
}


int __av_virtaccess(const char *path, int amode)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("access", path, amode, 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localaccess(&v, path, amode);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      res = __av_access(&v, vp, amode);
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (access)", path, res, v.errn);

  errno = v.errn;
  return res;
}

static int __av_readlink(ave *v, vpath *path, char *buf, avsize_t bsiz)
{
  int res;
  struct vdev_info *vdev = path->vdev;
  
  if(vdev == NULL) 
    res = __av_localreadlink(v, path->ipath, buf, bsiz);
  else {
    if(vdev->readlink != NULL)
      res = (*vdev->readlink)(v, vdev->devdata, path, buf, bsiz);
    else {
      v->errn = ENODEV;
      res = -1;
    }
  }

  return res;
}

int __av_virtreadlink(const char *path, void *buf, size_t bsiz)
{
  ave v;
  int res;
  int local;

  v.errn = errno;
  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localreadlink(&v, path, buf, bsiz);
    if(sure_local(&v, res, path)) local = 1;
  }

  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 0);
    if(vp != NULL) {
      res = __av_readlink(&v, vp, buf, bsiz);
      __av_free_vpath(vp);
    }
  }

  errno = v.errn;
  return res;
}

#ifdef HAVE_OPEN64

static LOCAL3(lseek64, offset_t, int, offset_t, int)

offset_t __av_virtlseek64(int fh, offset_t offset, int whence)
{
  ave v;
  offset_t res;

  v.errn = errno;

  AVLOGSYS("lseek64", (whence == SEEK_CUR) ? "SEEK_CUR" :
              ((whence == SEEK_SET) ? "SEEK_SET" : "SEEK_END"),
              fh, 0);
  AVLOGSYS("lseek64", "off low/hi",  (int) offset, (int) (offset >> 32));


  if(IS_LOCAL(fh) && !DO_VIRTDIR(fh)) 
    res = __av_locallseek64(&v, fh, offset, whence);
  else 
    res = __av_lseek(&v, fh, offset, whence);

  AVLOGSYS("  (lseek64)", "ret low/hi", (int) res, (int) (res >> 32));
  AVLOGSYS("  (lseek64)", "errno", v.errn, 0);

  errno = v.errn;
  return res;
}

#endif /* HAVE_OPEN64 */

#ifdef HAVE_ACL
#include <sys/acl.h>

static int __av_localacl(ave *v, const char *path, int cmd, 
		     int nent, aclent_t *aclbuf)
{
  int res;

  errno = v->errn;
  res = __av_realacl(path, cmd, nent, (void *) aclbuf);
  v->errn = errno;
  return res;
}


static int __av_acl(ave *v, vpath *path, int cmd,
		    int nent, aclent_t *aclbuf)
{
  struct vdev_info *vdev = path->vdev;
  int res;

  if(vdev == NULL)
    res = __av_localacl(v, path->ipath, cmd, nent, aclbuf);
  else {
    struct avstat stbuf;

    switch(cmd) {
    case GETACLCNT:
      res = 4;
      break;
      
    case SETACL:
      v->errn = ENOSYS;
      res = -1;
      break;

    case GETACL:
      res = __av_stat(v, path, &stbuf, AVFS_FLAG_NOSIZE);
      if(res != -1) {
	res = 0;
	if(nent > 0) {
	  aclbuf[0].a_type = USER_OBJ;
	  aclbuf[0].a_id   = stbuf.uid;
	  aclbuf[0].a_perm = (stbuf.mode & 0700) >> 6;
	  res = 1;
	}
	if(nent > 1) {
	  aclbuf[1].a_type = GROUP_OBJ;
	  aclbuf[1].a_id   = stbuf.gid;
	  aclbuf[1].a_perm = (stbuf.mode & 0070) >> 3;
	  res = 2;
	}
	if(nent > 2) {
	  aclbuf[2].a_type = CLASS_OBJ;
	  aclbuf[2].a_id   = -1;
	  aclbuf[2].a_perm = 0777;
	  res = 3;
	}
	if(nent > 3) {
	  aclbuf[3].a_type = OTHER_OBJ;
	  aclbuf[3].a_id   = -1;
	  aclbuf[3].a_perm = (stbuf.mode & 0007);
	  res = 4;
	}
      }
      break;

    default:
      v->errn = EINVAL;
      res = -1;
    }
  }

  return res;
}


int __av_virtacl(const char *path, int cmd, int nent, void *aclbuf)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("acl", path, cmd, nent);


  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localacl(&v, path, cmd, nent, aclbuf);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;
    
    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      res = __av_acl(&v, vp, cmd, nent, (aclent_t *) aclbuf);
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (acl)", path, res, v.errn);

  errno = v.errn;
  return res;
}

#endif /* HAVE_ACL */


