/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    OP_MISC module (vfs layer)
*/

#include "op_util.h"

#include <stdlib.h>
#include <unistd.h>

#define AVFSENVLEN 8

static char vcwdenv[AVFSENVLEN+AVFS_MAX_PATHLEN+1+1] =
{'A','V','F','S','_','C','W','D','=','\0'};

void __av_init_vcwd(void)
{
  char *avfs_cwd;

  __av_priv.vcwd_inited = 1;
  __av_priv.vcwd = vcwdenv + AVFSENVLEN + 1;

  avfs_cwd = getenv("AVFS_CWD");
  if(avfs_cwd != NULL) {
    strncpy(__av_priv.vcwd, avfs_cwd, AVFS_MAX_PATHLEN);
    __av_priv.vcwd[AVFS_MAX_PATHLEN] = '\0';
  }
  else __av_priv.vcwd[0] = '\0';
}

static LOCAL1(chdir, int, const char *)
static LOCAL1(fchdir, int, int)

static int __av_localexecve(ave *v, const char *filename, char *const argv[],
                        char *const envp[])
{
  int res;

  errno = v->errn;
  res = __av_realexecve(filename, argv, envp);
  v->errn = errno;
  return res;
}



static int __av_chdir(ave *v, vpath *path)
{
  struct vdev_info *vdev = path->vdev;
  struct avstat avstbuf;
  const char *localpath;
  char *rpath;
  int res;
  int origi, i;
  
  if(vdev == NULL) {
    res = __av_localchdir(v, path->ipath);
    if(res != -1) set_vcwd("");
    return res;
  }

  res = __av_stat(v, path, &avstbuf, AVFS_FLAG_NOSIZE);    
  if(res == -1) return -1;

  if(!AV_ISDIR(avstbuf.mode)) {
    v->errn = ENOTDIR;
    return -1;
  }

  origi = __av_generate_path(v, path, &rpath);
  if(origi == -1) return -1;
  
  i = origi - 1;

  /* If the local file ends in a / and it is a directory, then it will not
     become part of the virtual path */

  if(rpath[i] == DIR_SEP_CHAR) {
    vpath *locvp;

    for(locvp = path->base; locvp->vdev != NULL; locvp = locvp->base);
    res = __av_stat(v, locvp, &avstbuf, 0);
    if(res == -1) return -1;
    
    if(AV_ISDIR(avstbuf.mode)) i++;
  }

  while(i >= 0 && rpath[i] == DIR_SEP_CHAR) i--;
  if(i < 0) i = origi - 1;
  while(i >= 0 && rpath[i] != DIR_SEP_CHAR) i--;

  if(i < 0) localpath = "."; 
  else if(i == 0) localpath = "/";
  else {
    rpath[i] = '\0';
    localpath = rpath;
  }
  res = __av_localchdir(v, localpath);
  
  if(res != -1) set_vcwd(rpath+i+1);

  __av_free(rpath);

  return res;
}


int __av_virtchdir(const char *path)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  local = 0; res = -1;

  if(maybe_local(path)) {
    res = __av_localchdir(&v, path);
    if(res != -1) set_vcwd("");
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      res = __av_chdir(&v, vp);
      __av_free_vpath(vp);
    }
  }
  
  AVLOGSYS("chdir", path, res, v.errn);
  AVLOGSYS("  VCWD:", __av_priv.vcwd, 0, 0);
  
  errno = v.errn;
  return res;
}


/* Nobody uses fchdir for anything really useful, so I can't even properly
   test it! Arrgh... */

static int __av_fchdir(ave *v, int fh)
{
  finfo *fi;
  int res;

  if(IS_LOCAL(fh)) return __av_localfchdir(v, fh);

  fi = check_fh(v, fh);
  if(fi == NULL) return -1;

  if(fi->path == NULL) { /* should not happen */
    v->errn = EFAULT; 
    return -1;
  }

  res = __av_chdir(v, fi->path);
  
  return res;
}


int __av_virtfchdir(int fh)
{
  ave v;
  int res;

  v.errn = errno;
  res =  __av_fchdir(&v, fh);

  AVLOGSYS("fchdir", vcwdenv, res, v.errn);

  errno = v.errn;
  return res;
}

int __av_virtexecve(const char *filename, char *const argv[],
                    char *const envp[])
{
  int res;
  char **newep;
  char *savedenv;
  int i, at;
  ave v;

  v.errn = errno;

  if(!__av_priv.vcwd_inited) __av_init_vcwd();
  
  newep = NULL;
  at = -1;
  for(i = 0; envp[i] != NULL; i++)
    if(strncmp(envp[i], vcwdenv, AVFSENVLEN+1) == 0) {
      at = i;
      break;
    }

  AVLOGSYS("execve", filename, i, at);
  AVLOGSYS(__av_priv.vcwd, at != -1 ? envp[at] : "(null)",
              __av_priv.vcwd[0], at != -1 ? envp[at][AVFSENVLEN+1]: -1 );

/* Mallocing is unhealthy here, because tcsh redefines malloc, and
   here it does't work on solaris. So the user will have to set the
   AVFS_CWD env-variable beforehand, to make it work correctly */

/* FIXME: make own malloc functions, to be Safe Everywhere Always */

  savedenv = NULL;
  if(at != -1) {
    if(strcmp(envp[at], vcwdenv) != 0) {
      savedenv = envp[at];
      newep = (char **) envp; /* Stupid C compiler, I'm not doing
                                 anything bad, I'm really not! */
      newep[at] = vcwdenv;
    }
  }
  else if(__av_priv.vcwd[0]) {
#if 0
    AVLOGSYS("stupid user", "", 0, 0);
#endif
    at = i;
    i++;
    newep = __av_malloc(DUMMYV, sizeof(char *) * (i + 1));
    if(newep != NULL) {
      memcpy(newep, envp, sizeof(char *) * i);
      newep[at] = vcwdenv;
      newep[i] = NULL;
      res = __av_localexecve(&v, filename, argv, newep);
      __av_free(newep);
      errno = v.errn;
      return res;
    }
  }

  res = __av_localexecve(&v, filename, argv, envp);
  if(savedenv != NULL) newep[at] = savedenv;
  errno = v.errn;
  return res;
}


#define COPY_BUFSIZE 16384

static avssize_t copy_file(ave *v, int destfd, int srcfd, avoff_t offset, 
			   avoff_t len)
{
  char buf[COPY_BUFSIZE];
  avoff_t currpos;
  avssize_t ret;
  avssize_t written;

  currpos = __av_lseek(v, srcfd, 0, SEEK_CUR);
  if(currpos == -1) return -1;

  if(__av_lseek(v, srcfd, offset, SEEK_SET) == -1) return -1;
  
  written = 0;
  do {
    ret = __av_read(v, srcfd, buf, COPY_BUFSIZE);
    if(ret == -1) return -1;
    if(ret > 0) {
      if(__av_localwrite(v, destfd, buf, ret) == -1) return -1;
      written += ret;
    }
  } while(ret > 0 && (len < 0 || written < len));

  __av_lseek(v, srcfd, currpos, SEEK_SET);

  return written;
}

real_file *__av_get_realfile(ave *v, vpath *path)
{
  int srcfd, destfd;
  int res = -1;
  real_file *rf;

  AV_NEW(v, rf);
  if(rf == AVNULL) return AVNULL;

  if(path->vdev == AVNULL) {
    rf->name = __av_strdup(v, path->ipath);
    if(rf->name == AVNULL) {
      __av_free(rf);
      return AVNULL;
    }
    rf->is_tmp = 0;
    return rf;
  }

  rf->name = __av_get_tmpfile(v);
  if(rf->name == AVNULL) {
    __av_free(rf);
    return AVNULL;
  }
  rf->is_tmp = 1;
  
  srcfd = __av_open(v, path, AVO_RDONLY, 0);
  if(srcfd != -1) {
    destfd = __av_localopen(v, rf->name, AVO_WRONLY | AVO_CREAT | AVO_TRUNC, 
			    0600);
    if(destfd != -1) {
      res = copy_file(v, destfd, srcfd, 0, -1);
      __av_localclose(DUMMYV, destfd);
    }
    __av_close(DUMMYV, srcfd);
  }
  
  if(res == -1) {
    __av_free_realfile(rf);
    return AVNULL;
  }

  return rf;
}

void __av_free_realfile(real_file *rf)
{
  if(rf != AVNULL) {
    if(!rf->is_tmp) 
      __av_free(rf->name);
    else 
      __av_del_tmpfile(rf->name);
    
    __av_free(rf);
  }
}



int __av_fgetfile(ave *v, int fh, avoff_t offset, avsize_t *lenp,
		 avoff_t *retoffp)
{
  finfo *fi;
  int newfd;

  AVDEB((AVLOG_DEBUG, "fgetfile: %i off: %i, len: %i", fh, offset, *lenp));

  if(IS_LOCAL(fh)) {
    *retoffp = offset;
    newfd = dup(fh);

    AVDEB((AVLOG_DEBUG, "  (fgetfile): local %i newoff: %i, newlen: %i [%i]", 
	 newfd, *retoffp, *lenp, v->errn));
    return newfd;
  }

  fi = check_fh(v, fh);
  if(fi == NULL) return -1;

  if(fi->vdev->fgetfile != NULL) 
    newfd = (*fi->vdev->fgetfile) (v, fi->devinfo, offset, lenp, retoffp);
  else {
    v->errn = ENODEV;
    newfd = -1;
  }

  AVDEB((AVLOG_DEBUG, "  (fgetfile): %i newoff: %i, newlen: %i [%i]", 
	   newfd, *retoffp, *lenp, v->errn));
    
  return newfd;
}

#ifdef HAVE_MMAP
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>

static LOCAL6(mmap, void *, void *, size_t, int, int, int, off_t)

static void *__av_mmap(ave *v, void *start, size_t loclen, int prot,
                        int flags, int fd, off_t offset)
{
  int mmapfd;
  void *res;
  avsize_t len = loclen;
  avoff_t realoff;
  
  if((prot & PROT_WRITE) && (flags & MAP_SHARED)) {
    v->errn = EACCES;
    return (void *) -1;
  }
  
  mmapfd = __av_fgetfile(v, fd, offset, &len, &realoff);
  if(mmapfd == -1 && v->errn == ENODEV) {
    char *dummyfile;
    avssize_t written;
  
    dummyfile = __av_get_tmpfile(v);
    if(dummyfile == AVNULL) mmapfd = -1;
    else {
      mmapfd = __av_localopen(v, dummyfile, AVO_RDWR | AVO_CREAT | AVO_TRUNC, 
			      0600);
      
      __av_del_tmpfile(dummyfile);
      
      if(mmapfd != -1) {
      
	/* FIXME: lock fh */
	written = copy_file(v, mmapfd, fd, offset, len);
	if(written == -1) {
	  __av_localclose(DUMMYV, mmapfd);
	  mmapfd = -1;
	}
	else {
	  realoff = 0;
	  len = AV_MIN(len, (avsize_t) written);
	}
      }
    }
  }

  if(mmapfd == -1) return (void *) -1;

  res = __av_localmmap(v, start, len, prot, flags, mmapfd, realoff);
  __av_localclose(DUMMYV, mmapfd);

  return res;
}

void *__av_virtmmap(void *start, size_t len, int prot,
                           int flags, int fd, off_t offset)
{
  ave v;
  void *res;

  v.errn = errno;

  AVLOGSYS("mmap", "fd/start", fd, (int) start);
  AVLOGSYS("mmap", "prot/offset", prot, offset);

  if(IS_LOCAL(fd)) 
    res = __av_localmmap(&v, start, len, prot, flags, fd, offset);
  else 
    res = __av_mmap(&v, start, len, prot, flags, fd, offset);

  AVLOGSYS("  (mmap)", "", (int) res, v.errn);

  errno = v.errn;
  return res;
}

#endif /* HAVE_MMAP */


#ifdef HAVE_OPEN64

static LOCAL6(mmap64, void *, void *, size_t, int, int, int, offset_t)

void *__av_virtmmap64(void *start, size_t len, int prot,
		      int flags, int fd, offset_t offset)
{
  ave v;
  void *res;

  v.errn = errno;

  AVLOGSYS("mmap64", "", fd, (int) start);
  AVLOGSYS("mmap64", "off low/hi",  (int) offset, (int) (offset >> 32));

  if(IS_LOCAL(fd)) 
    res =  __av_localmmap64(&v, start, len, prot, flags, fd, offset);
  else 
    res = __av_mmap(&v, start, len, prot, flags, fd, (int) offset);

  AVLOGSYS("  (mmap64)", "", (int) res, v.errn);

  errno = v.errn;
  return res;
}


#endif

