/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    OP_OPEN module (vfs layer)
*/

#include "op_util.h"

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>


LOCAL1(close, int, int)

static int loc_open(ave *v, const char *path, int flags, int mode)
{
  int res;

  errno = v->errn;
  res = __av_realopen(path, flags, mode);
  v->errn = errno;
  
  return res;
}

static int get_handle(ave *v)
{
  int fh;
  char dummyfile[64];

  /* FIXME: Is this OK?  /dev/null and such are no good, because they
     block select(2) */

  strcpy(dummyfile, "/tmp/.avfs_dummyfile_XXXXXX");
  mktemp(dummyfile);
  if(dummyfile[0] == '\0') {
    v->errn = ENFILE;   /* FIXME: These error values should be documented */
    return -1;
  }

  /* FIXME: Does it need to be opened for writing? */
  fh = __av_realopen(dummyfile, O_RDONLY | O_CREAT | O_EXCL, 0666);
  __av_realunlink(dummyfile);

  if(fh == -1) {
    v->errn = errno;
    return -1;
  }
  
  if(fh < 3) {
    /* If file descriptor is too small: (0,1,2), then better get
       another one! */
    
    int nfh;

    nfh = fcntl(fh, F_DUPFD, 3);
    if(nfh == -1) {
      v->errn = errno;
      __av_realclose(fh);
      return -1;
    }
    __av_realclose(fh);
    fh = nfh;
  }

  if(!FD_OK(fh)) {
    __av_realclose(fh);
    v->errn = ENFILE;
    return -1;
  }

  if(ISVIRTUAL(fh)) {
    __av_realclose(fh);
    __av_priv.file_table[fh].type = 0; /* reset offending fh */
    v->errn = EFAULT;        /* meaning internal error */
    return -1;
  }

  fcntl(fh, F_SETFD, FD_CLOEXEC);  /* Virtual files do not remain open
				      accross exec, so we might as
				      well close the dummy fd.  */

  __av_priv.file_table[fh].type = FVIRTUAL;
  return fh;

}

static void free_handle(int fh)
{
  __av_realclose(fh);
  if(FD_OK(fh)) __av_priv.file_table[fh].type = 0;
}


/* This function checks for the necessity of "virtual mountpoints" in
   directory listing. This is needed for pwd-like routines */
static void check_virtual_dir(int fh, vpath *path)
{
  int needvirt;
  int vci = 0;
  int res;
  vpath *vp;
  int i;

#if 0
  AVLOGSYS("check_virtual_dir", path->ipath, fh, 0);
#endif

  if(!__av_priv.vcwd[0] || !FD_OK(fh)) return;

  /* Only relative paths are considered */
  for(vp = path; vp->vdev != NULL; vp = vp->base);
  if(vp->ipath[0] == DIR_SEP_CHAR) return;

  needvirt = 0;

  if(path->vdev == NULL && path->ipath[0] == '.' && 
     (!path->ipath[1] || 
      (path->ipath[1] == DIR_SEP_CHAR && !path->ipath[2]))) {
    needvirt = 1;
    vci = 0;
  }
  else {
    char *rpath;
    int plen;
    
    res = __av_generate_path(DUMMYV, path, &rpath);
    if(res == -1) return;
    
    plen = strlen(rpath);
    
    if(strncmp(__av_priv.vcwd, rpath, plen) == 0 && 
       __av_priv.vcwd[plen] == DIR_SEP_CHAR) {
      vci = plen+1;
      for(i = vci; __av_priv.vcwd[i] && __av_priv.vcwd[i] != DIR_SEP_CHAR; i++)
	if(__av_priv.vcwd[i] == VDEV_SEP_CHAR) {
	  needvirt = 1;
	  break;
	}
    }
     __av_free(rpath);
  }

  if(needvirt) {
    struct avdirent *ent;

    AV_NEW(DUMMYV, ent);
    if(ent != NULL) {
      for(i = vci; __av_priv.vcwd[i] && 
	    __av_priv.vcwd[i] != DIR_SEP_CHAR; i++);

      ent->reclen = VIRTDIRSIZE;
      ent->off = -1;
      strncpy(ent->name, __av_priv.vcwd+vci, i-vci);
      ent->name[i - vci] = '\0';
      ent->ino = 1; /* FIXME: Ugly constant */

      __av_priv.file_table[fh].virtdir = 1;
      __av_priv.file_table[fh].virtent = ent;

#if 0
      AVLOGSYS("VIRDIR OK", path->ipath, fh, 0);
      AVLOGSYS("virtent", ent->name, ent->ino, 0);
#endif
    }
  }
}


static int common_open(ave *v, vpath *path, int avflags, int mode)
{
  int fh;
  void *devinfo;
  struct vdev_info *vdev = path->vdev;

  if(vdev->open == NULL) {
    v->errn = ENODEV;
    return -1;
  }

  fh = get_handle(v);
  if(fh != -1) {
    devinfo = (*vdev->open)(v, vdev->devdata, path, avflags, mode);
    if(devinfo == NULL) {
      free_handle(fh);
      fh = -1;
    }
    else {
      __av_priv.file_table[fh].devinfo  = devinfo;
      __av_priv.file_table[fh].vdev     = vdev;
      __av_priv.file_table[fh].path     = NULL;
      __av_priv.file_table[fh].oflags   = avflags;
    }
  }
  return fh;
}


static int convert_oflags(int flags)
{
  int avflags;
  
  avflags = flags & 3;
  if(flags & O_CREAT)  avflags |= AVO_CREAT;
  if(flags & O_EXCL)   avflags |= AVO_EXCL;
  if(flags & O_TRUNC)  avflags |= AVO_TRUNC;
  if(flags & O_APPEND) avflags |= AVO_APPEND;

  return avflags;
}

static int convert_oflags_back(int avflags)
{
  int flags;
  
  flags = avflags & AVO_ACCMODE;
  if(avflags & AVO_CREAT)  flags |= O_CREAT;
  if(avflags & AVO_EXCL)   flags |= O_EXCL;
  if(avflags & AVO_TRUNC)  flags |= O_TRUNC;
  if(avflags & AVO_APPEND) flags |= O_APPEND;

  return flags;
}

int __av_localopen(ave *v, const char *path, int avflags, int mode)
{
  return loc_open(v, path, convert_oflags_back(avflags), mode);
}


int __av_open(ave *v, vpath *path, int avflags, int mode)
{
  int fh;

  if(path->vdev == NULL) 
    fh = __av_localopen(v, path->ipath, avflags, mode);
  else
    fh = common_open(v, path, avflags, mode);

  return fh;
}

static int do_open(ave *v, const char *path, int flags, int mode)
{
  vpath *vp;
  int fh;
  
  vp = __av_get_vpath(v, path, 1);
  if(vp != NULL) {
    if(vp->vdev == NULL) 
      fh = loc_open(v, vp->ipath, flags, mode);
    else
      fh = common_open(v, vp, convert_oflags(flags), mode);
    
    check_virtual_dir(fh, vp);
    
    if(!IS_LOCAL(fh)) __av_priv.file_table[fh].path = vp;
    else __av_free_vpath(vp);
  }
  else fh = -1;

  return fh;
}

int __av_virtopen(const char *path, int flags, int mode)
{
  ave v;
  int fh;
  int local;
  struct stat stbuf;
  int pathloc = 0;

  v.errn = errno;

  AVLOGSYS("open", path, flags, mode);

  local = 0; fh = -1;
  if(maybe_local(path) && (!(flags & O_CREAT) || 
			   (pathloc = path_local(path)) || 
			   __av_realstat(path, &stbuf) != -1)) {
    fh = loc_open(&v, path, flags, mode);
    if(pathloc || sure_local(&v, fh, path)) local = 1;
  }

  if(!local) fh = do_open(&v, path, flags, mode);
    
  AVLOGSYS("  (open)", path, fh, v.errn);

  errno = v.errn;
  return fh;
}

int __av_virtcreat(const char *path, mode_t mode)
{
  return __av_virtopen(path, O_WRONLY | O_CREAT | O_TRUNC, mode);
}

int __av_close(ave *v, int fh)
{
  finfo *fi;
  int res;

  if(DO_VIRTDIR(fh)) {
    __av_priv.file_table[fh].virtdir = 0;
    __av_free(__av_priv.file_table[fh].virtent);
  }
  
  if(IS_LOCAL(fh)) return __av_localclose(v, fh);

  fi = check_fh(v, fh);
  if(fi == NULL) return -1;

  res = (*fi->vdev->close)(v, fi->devinfo);

  fi->devinfo = NULL;
  free_handle(fh);

  return res;
}

int __av_virtclose(int fh)
{
  ave v;
  int res;
  vpath *vp = NULL;

  v.errn = errno;
  if(!IS_LOCAL(fh)) vp = __av_priv.file_table[fh].path;
  res = __av_close(&v, fh);
  if(vp != NULL) __av_free_vpath(vp);
  errno = v.errn;
  return res;
}

#ifdef HAVE_OPEN64

static LOCAL3(open64, int, const char *, int, int)

int __av_virtopen64(const char *path, int flags, int mode)
{
  ave v;
  int fh;
  int local;
  struct stat stbuf;
  int pathloc = 0;

  v.errn = errno;

  AVLOGSYS("open64", path, flags, mode);

  local = 0; fh = -1;
  if(maybe_local(path) && (!(flags & O_CREAT) || 
			   (pathloc = path_local(path)) || 
			   __av_realstat(path, &stbuf) != -1)) {
    fh = __av_localopen64(&v, path, flags, mode);
    if(pathloc || sure_local(&v, fh, path)) local = 1;
  }

  if(!local) fh = do_open(&v, path, flags, mode);
    
  AVLOGSYS("  (open64)", path, fh, v.errn);

  errno = v.errn;
  return fh;
}

int __av_virtcreat64(const char *path, mode_t mode)
{
  return __av_virtopen64(path, O_WRONLY | O_CREAT | O_TRUNC, mode);
}

#endif /* HAVE_OPEN64 */


#ifdef EMULATE_DIROPS

static LOCAL1(opendir, DIR *, const char *)
static LOCAL1(closedir, int, DIR *)

static AVFS_DIR *__av_opendir(ave *v, vpath *path)
{
  AVFS_DIR *di;
  struct dirent *ent;
  int success;

  if(path->vdev == NULL) {
    /* This hack is needed, because opendir() calls stat()... */
    /* FIXME: Really need to lock this region... */

    char vcwdback[AVFS_MAX_PATHLEN+1];

    strcpy(vcwdback, __av_priv.vcwd);
    set_vcwd("");
    di = (AVFS_DIR *) __av_localopendir(v, path->ipath);
    set_vcwd(vcwdback);
  }
  else {
    success = 0;
    AV_NEW(v, di);
    ent = (struct dirent *)
      __av_malloc(v, sizeof(struct dirent) + AVFS_MAX_NAMELEN + 1);

    if(di != NULL && ent != NULL) {

      di->fd = common_open(v, path, 0, 0);
      di->buf = (char *) ent;
      if(di->fd != -1) success = 1;
    }

    if(!success) {
      __av_free(ent);
      __av_free(di);
      di = NULL;
    }
  }

  if(di != NULL) check_virtual_dir(di->fd, path);

  return (AVFS_DIR *) di;
}

void *__av_virtopendir(const char *path)
{
  ave v;
  AVFS_DIR *dirp;
  int local;

  v.errn = errno;

  local = 0; dirp = NULL;
  if(maybe_local(path)) {
    dirp = (AVFS_DIR *) __av_localopendir(&v, path);
    if(sure_local(&v, (dirp == NULL) ? -1 : 0, path)) local = 1;
  }

  if(!local) {
    vpath *vp;
    
    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      dirp = __av_opendir(&v, vp);
      if(dirp != NULL && !IS_LOCAL(dirp->fd)) 
	__av_priv.file_table[dirp->fd].path = vp;
      else 
	__av_free_vpath(vp);
    }
  }

  AVLOGSYS("opendir", path, dirp != NULL ? dirp->fd : -1, v.errn);

  errno = v.errn;
  return (void *) dirp;
}

static int __av_closedir(ave *v, AVFS_DIR *dirp)
{
  int res;

  if(DO_VIRTDIR(dirp->fd)) {
    __av_priv.file_table[dirp->fd].virtdir = 0;
    __av_free(__av_priv.file_table[dirp->fd].virtent);
  }

  if(IS_LOCAL(dirp->fd)) return __av_localclosedir(v, (DIR *) dirp);

  res = __av_close(v, dirp->fd);
  __av_free(dirp->buf);
  __av_free(dirp);

  return res;
}

int __av_virtclosedir(void *dirp)
{
  ave v;
  int res;
  AVFS_DIR *avdirp = (AVFS_DIR *) dirp;
  vpath *vp = NULL;

  v.errn = errno;
  if(!IS_LOCAL(avdirp->fd)) vp = __av_priv.file_table[avdirp->fd].path;
  res = __av_closedir(&v, avdirp);
  if(vp != NULL) __av_free_vpath(vp);
  errno = v.errn;
  return res;
}

#endif /* EMULATE_DIROPS */
